/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvio.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVIO_H
#define QVIO_H

#include <qvdefines.h>
#include <QVImage>
#include <QFile>

/*!
@brief Create a yuv4mpeg2 video file.

This function stores the header for a video sequence in YUV4MPEG2 format in a newly created file.
It can be used in conjunction with any of the versions of the @ref writeYUV4MPEG2Frame functions to store
a flow of images in a video file.

The following code illustrates its usage. First the video file must be created and initializated with a header:

@code
#include <QFile>

[...]
QFile videoFile("test.mpeg");
videoFile.open(QIODevice::WriteOnly|QIODevice::Truncate);
writeYUV4MPEG2Header(videoFile, cols, rows, 24);
[...]
@endcode

Then, each image frame must be stored in the video file, until a stop condition:

@code
while(newImages)
	{
	// Read a new image in RGB format
	QVImage<uChar,3> rgbImage = getNewRGBImage();
	
	// Write the RGB image to the file
	writeYUV4MPEG2Frame(videoFile, rgbImage);
	[...]
	}
@endcode

Alternatively, an overloaded version of the @ref writeYUV4MPEG2Frame function can be used, to store an image in YUV format:

@code
while(newImages)
	{
	// Read a new YUV image in 4:2:0 format
	QVImage<uChar,1> yImage = getNewYUVImageYChannel();
	QVImage<uChar,1> UImage = getNewYUVImageUChannel();
	QVImage<uChar,1> VImage = getNewYUVImageVChannel();

	// Write the YUV image to the file
	writeYUV4MPEG2Frame(videoFile, yImage, uImage, vImage);
	[...]
	}
@endcode

when finished storing frames in the video file, it can be closed

@code
videoFile.close();
@endcode

@param file The input file object. It must be already opened when received as argument by this function.
@param cols The number of columns that the video in the file should have.
@param rows The number of rows that the video in the file should have.
@param fps The number of frames per second that the video sequence will have.
@see writeYUV4MPEG2Frame(QFile &, const QVImage<uChar,1>, const QVImage<uChar,1>, const QVImage<uChar,1>);
@see writeYUV4MPEG2Frame(QFile &, const QVImage<uChar,3>);
@ingroup qvio
*/
bool writeYUV4MPEG2Header(QFile &file, const int cols, const int rows, const int fps);

/*!
@brief Write YUV image frames to a yuv4mpeg2 video file.

This is an overloaded version of the function  @ref writeYUV4MPEG2Frame(QFile &, const QVImage<uChar,3>) provided for convenience.

This version stores a frame in a video file from the YUV format. The parameters of this function must be three images containing the
three chanels for the YUV image.
For further information about its usage, see documentation for function @ref writeYUV4MPEG2Header.


@param file The input file object. It must be already opened when received as argument by this function.
@param imageY Image object containing the Y channel for the YUV image, to store in the video file.
@param imageU Image object containing the Y channel for the YUV image, to store in the video file.
@param imageV Image object containing the Y channel for the YUV image, to store in the video file.
@see writeYUV4MPEG2Header
@ingroup qvio
*/
bool writeYUV4MPEG2Frame(QFile &file, const QVImage<uChar,1> imageY, const QVImage<uChar,1> imageU, const QVImage<uChar,1> imageV);

/*!
@brief Write RGB image frames to a yuv4mpeg2 video file.

This function stores a frame in a video file from the RGB format. The parameter of this function is an image containing the RGB image.
For further information about its usage, see documentation for function @ref writeYUV4MPEG2Header.

@param file The input file object. It must be already opened when received as argument by this function.
@param imageRGB RGB image object, to store in the video file.
@see writeYUV4MPEG2Header
@ingroup qvio
*/
bool writeYUV4MPEG2Frame(QFile &file, const QVImage<uChar,3> imageRGB);

/*!
@brief Write the content of a QVImage object in an image file

@todo Document this

@param fileName The name of the image file to create. It is overwritten if already exists.
                File format is inferred from fileName suffix. Supported formats:
                BMP,JPG,PNG,PPM,TIFF,XBM,XPM.
@param image The image to store in the file.
@returns TRUE if success, FALSE otherwise.
@ingroup qvio
*/
bool writeQVImageToFile(const QString fileName, const QVImage<uChar, 3> &image);

/*!
@brief Reads the content of a QVImage object from an image file

@todo Document this

@param fileName The name of the image file to read from. File format is inferred from fileName suffix.
                Supported formats: BMP,JPG,PNG,PPM,TIFF,XBM,XPM,GIF,PBM,PGM.
@param image The image to store in the file.
@returns TRUE if success, FALSE otherwise.
@ingroup qvio
*/
bool readQVImageFromFile(const QString fileName, QVImage<uChar, 3> &image);

#endif
