/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvvideorecorder.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVVIDEORECORDER_H
#define QVVIDEORECORDER_H

#include <QVWorker>
#include <qvio.h>

/*!
@class QVVideoRecorder qvvideorecorder.h qvio/qvvideorecorder.h
@brief Video source class for reading many video file formats, webcams and digital cameras.

This worker can be linked to a worker object producing an output image, to store the flow of images in a video file in the uncompressed YUV4MPEG2 format.

To use it, create a QVVideoRecorder object in the <i>main function</i>, and link a RGB image from a worker, to store each output image from the worker in a video file.

@code
int main(int argc, char *argv[])
	{
	[...]
	// The following worker should have a property named 'Output image' containing an RGB image.
	MyWorker myWorker("My worker");
	[...]
	QVVideoRecorder recorder("Video recorder");
	recorder.linkProperty(player, "Output image");
	[...]
	}
@endcode

Alternatively you can link three properties containing the Y, U and V channels of a YUV image respectively to store  images in the video file directly in that format:

@code
int main(int argc, char *argv[])
	{
	[...]
	// The following worker should have three properties named 'Output image Y', 'Output image U' and
	// 'Output image V' representing the three channels of a YUV image.
	MyWorker myWorker("My worker");
	[...]
	QVVideoRecorder recorder("Video recorder");
	recorder.linkProperty(player, "Output image Y", "Output image U", "Output image V");
	[...]
	}
@endcode

You can tune the number of frames per second and the name of the file to store the video using functions @ref setFPS and @ref setVideoFileName respectively, or through the user command line, setting values for the properties <i>FPS</i> and <i>Record file name</i> in the video record object.

Once the object reads the first image frame changing those values will have no effect.

@ingroup qvio
*/
class QVVideoRecorder: public QVWorker
	{
	private:
		bool initiated, rgbMode;
		QFile videoFile;

		bool linkProperty(QString sourcePropertyName, QVPropertyContainer *destinyContainer, QString destinyPropertyName, LinkType linkType);

	public:
		QVVideoRecorder(QString name, const QString fileName = "pru.mpg", const int fps = 25);
		~QVVideoRecorder();

		/// @todo Check image types
		bool linkProperty(QVWorker &worker, const QString rgbImageName);

		/// @todo Check image types
		bool linkProperty(QVWorker &worker, const QString yImageName, const QString uImageName, const QString vImageName);

		void iterate();

		void setFPS(const int fps)				{ setPropertyValue<int>("FPS", fps); }

		void setVideoFileName(const QString &videoFileName)	{ setPropertyValue<QString>("Record file name", videoFileName); }
	};

#endif
